#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#ifdef SOLARIS
#  include <strings.h>
#endif
#include "lcd.h"
#include "report.h"
#include "lcd_lib.h"
#include "adv_bignum.h"

#define JBC1602_DEFAULT_DEVICE "/dev/ttyUSB0"

#define JBC1602_LINE_WIDTH 16
#define JBC1602_NUM_LINES 2

/** private data for the \c jbc1602 driver */
typedef struct jbc1602_private_data {
	char device[256];
	int speed;				/* Bxxxx speed definition */
	int fd;					/* File descriptor to tty device */

	/* Framebuffer for device */
	unsigned char framebuf[JBC1602_NUM_LINES][JBC1602_LINE_WIDTH + 1];
	unsigned char shadow[JBC1602_NUM_LINES][JBC1602_LINE_WIDTH + 1];

	/* Input buffer for device */
	char input[64];
} PrivateData;


// Vars for the server core
MODULE_EXPORT char * api_version = API_VERSION;
MODULE_EXPORT int stay_in_foreground = 0;
MODULE_EXPORT int supports_multiple = 0;
MODULE_EXPORT char *symbol_prefix = "jbc1602_";


MODULE_EXPORT void jbc1602_flush(Driver *drvthis);
MODULE_EXPORT void jbc1602_clear(Driver *drvthis);

/**
 * Initialize the driver.
 * \param drvthis  Pointer to driver structure.
 * \retval 0       Success.
 * \retval <0      Error.
 */
MODULE_EXPORT int
jbc1602_init(Driver *drvthis)
{
	PrivateData *p;
	struct termios portset;
	unsigned char dummy[16];

	/* Allocate and store private data */
	p = (PrivateData *) calloc(1, sizeof(PrivateData));
	if (p == NULL)
		return -1;
	if (drvthis->store_private_ptr(drvthis, p))
		return -1;

	/* initialize private data */
	p->fd = -1;
	p->speed = B9600;
	jbc1602_clear(drvthis);

	/* Read config file */

	/* What device should be used */
	strncpy(p->device, drvthis->config_get_string(drvthis->name, "Device", 0,
		JBC1602_DEFAULT_DEVICE), sizeof(p->device));
	p->device[sizeof(p->device)-1] = '\0';
	report(RPT_INFO, "%s: using Device %s", drvthis->name, p->device);

	/* What speed to use */
	p->speed = drvthis->config_get_int(drvthis->name, "Speed", 0, 9600);

	if (p->speed == 1200)       p->speed = B1200;
	else if (p->speed == 2400)  p->speed = B2400;
	else if (p->speed == 9600)  p->speed = B9600;
	else if (p->speed == 19200) p->speed = B19200;
	else if (p->speed == 38400) p->speed = B38400;
	else if (p->speed == 57600) p->speed = B57600;
	else if (p->speed == 115200) p->speed = B115200;
	else {
		report(RPT_WARNING, "%s: illegal Speed %d; using default %d",
			drvthis->name, p->speed, 9600);
		p->speed = B9600;
	}

	// Set up io port correctly, and open it...
	p->fd = open(p->device, O_RDWR | O_NOCTTY | O_NDELAY);
	if (p->fd < 0) {
		report(RPT_ERR, "%s: open(%s) failed (%s)", drvthis->name, p->device, strerror(errno));
		return -1;
	}

	report(RPT_DEBUG, "jbc1602_init: opened device %s fd=%d", p->device, p->fd);

	tcflush(p->fd, TCIOFLUSH);

	tcgetattr(p->fd, &portset);
	portset.c_cflag &= ~(CSIZE | PARENB) | (CS8);
	portset.c_iflag &= ~(INPCK | CSTOPB);
	portset.c_lflag  &= ~(ICANON | ECHO | ECHOE | ISIG);  /*Input*/
	portset.c_cflag &= ~CRTSCTS;
	portset.c_oflag  &= ~OPOST;   /*Output*/
	portset.c_cc[VTIME] = 30;
	portset.c_cc[VMIN] = 0;

	// Set port speed
	cfsetospeed(&portset, p->speed);
	cfsetispeed(&portset, p->speed);

	// Do it...
	tcsetattr(p->fd, TCSANOW, &portset);
	tcflush(p->fd, TCIOFLUSH);

	/* Write a bunch of zeros to get in sync with the device */
	memset(dummy, 0, sizeof(dummy));
	write(p->fd, dummy, sizeof(dummy));

	/* Clear the JBC1602 */
	jbc1602_flush(drvthis);

	report(RPT_DEBUG, "%s: init() done", drvthis->name);

	return 0;
}


/**
 * Close the driver (do necessary clean-up).
 * \param drvthis  Pointer to driver structure.
 */
MODULE_EXPORT void
jbc1602_close(Driver *drvthis)
{
	PrivateData *p = drvthis->private_data;

	debug(RPT_DEBUG, "Closing JBC1602");
	if (p != NULL) {
		if (p->fd >= 0) {
			struct timespec ts;

			ts.tv_sec = 0;
			ts.tv_nsec = 200 * 1000000; // 200 ms
			tcdrain(p->fd);
			while (-1 == nanosleep(&ts, &ts));

			close(p->fd);
		}

		free(p);
	}
	drvthis->store_private_ptr(drvthis, NULL);
}

/**
 * Return the display width in characters.
 * \param drvthis  Pointer to driver structure.
 * \return         Number of characters the display is wide.
 */
MODULE_EXPORT int
jbc1602_width(Driver *drvthis)
{
	return JBC1602_LINE_WIDTH;
}


/**
 * Return the display height in characters.
 * \param drvthis  Pointer to driver structure.
 * \return         Number of characters the display is high.
 */
MODULE_EXPORT int
jbc1602_height(Driver *drvthis)
{
	return JBC1602_NUM_LINES;
}


/**
 * Clear the screen.
 * \param drvthis  Pointer to driver structure.
 */
MODULE_EXPORT void
jbc1602_clear(Driver *drvthis)
{
	PrivateData *p = drvthis->private_data;
	int i;

	for (i = 0; i < JBC1602_NUM_LINES; i++) {
		memset(p->framebuf[i], ' ', JBC1602_LINE_WIDTH);
	}
}


/**
 * Flush data on screen to the LCD.
 * \param drvthis  Pointer to driver structure.
 */
MODULE_EXPORT void
jbc1602_flush(Driver *drvthis)
{
	int i, j;
	int r;
	PrivateData *p = drvthis->private_data;
	unsigned char * b;

	debug(RPT_DEBUG, "JBC1602 flush");
	for (i = 0; i < JBC1602_NUM_LINES; i++)
	if (memcmp(p->shadow[i], p->framebuf[i], JBC1602_LINE_WIDTH + 1)) {
		p->framebuf[i][0] = 0x10 + i;
		b = (unsigned char *)(p->framebuf[i]);
		j = JBC1602_LINE_WIDTH + 1;

		while (j > 0) {
			fd_set brfdset;
			fd_set bwfdset;
		
			FD_ZERO(&brfdset);
			FD_ZERO(&bwfdset);
			FD_SET(p->fd, &brfdset);
			FD_SET(p->fd, &bwfdset);
			if (select(p->fd + 1, &brfdset, &bwfdset, NULL, NULL) >= 0) {
				if (FD_ISSET(p->fd, &brfdset)) {
					int bufused, remspace, r;

					bufused = strlen(p->input);
					remspace = sizeof(p->input) - bufused - 1;
					if (remspace <= 0) {
						report(RPT_ERR, "%s: Read overflow in JBC1602 flush", drvthis->name);
						memset(p->input, 0, sizeof(p->input));
						remspace = sizeof(p->input) - 1;
						bufused = 0;
					}
					r = read(p->fd, p->input + bufused, remspace);
					if (r < 0 && errno != EAGAIN) {
						report(RPT_ERR, "Failed to fetch ready data on fd=%d while flushing framebuffer", p->fd);
					}
				}
				if (FD_ISSET(p->fd, &bwfdset)) {
					r = write(p->fd, b, j);
					if (r > 0) {
						b += r;
						j -= r;
					} else {
						report(RPT_ERR, "Failed framebuffer flush on fd=%d", p->fd);
					}
				}
			}
		}
		tcdrain(p->fd);
		memcpy(p->shadow[i], p->framebuf[i], JBC1602_LINE_WIDTH + 1);
	}
}


/**
 * Print a string on the screen at position (x,y).
 * The upper-left corner is (1,1), the lower-right corner is (p->width, p->height).
 * \param drvthis  Pointer to driver structure.
 * \param x        Horizontal character position (column).
 * \param y        Vertical character position (row).
 * \param string   String that gets written.
 */
MODULE_EXPORT void
jbc1602_string(Driver *drvthis, int x, int y, const char string[])
{
	PrivateData *p = drvthis->private_data;
	int i;

	debug(RPT_DEBUG, "Putting string %s at %i, %i", string, x, y);

	x--;  // Convert 1-based coords to 0-based...
	y--;

	for (i = 0; string[i] != '\0'; i++) {
		unsigned char c = (unsigned char) string[i];

		// Check for buffer overflows...
		if (y >= JBC1602_NUM_LINES) break;
		if (x + i >= JBC1602_LINE_WIDTH) break;
#if 0
		if ((c > 0x7F) && (c < 0x98)) {
			report(RPT_WARNING, "%s: illegal char 0x%02X requested in jbc1602_string()",
				drvthis->name, c);
			c = ' ';
		}
#endif
#if 0
		if (c < 8)      /* The custom characters are mapped at 0x98 - 0x9F, */
			c += 0x98;   /* as 0x07 makes a beep instead of printing a character */
#endif
		p->framebuf[y][1 + x + i] = c;
  }
}

/**
 * Print a character on the screen at position (x,y).
 * The upper-left corner is (1,1), the lower-right corner is (p->width, p->height).
 * \param drvthis  Pointer to driver structure.
 * \param x        Horizontal character position (column).
 * \param y        Vertical character position (row).
 * \param c        Character that gets written.
 */
MODULE_EXPORT void
jbc1602_chr(Driver *drvthis, int x, int y, char c)
{
	PrivateData *p = drvthis->private_data;
	unsigned char ch = (unsigned char) c;

	debug(RPT_DEBUG, "Putting char %c (%#x) at %i, %i", c, c, x, y);

	y--;
	x--;
#if 0
	if ((ch > 0x7F) && (ch < 0x98)) {
		report(RPT_WARNING, "%s: illegal char 0x%02X requested in jbc1602_chr()",
			drvthis->name, c);
		ch = ' ';
	}
#endif
	if (y >= JBC1602_NUM_LINES) return;
	if (x >= JBC1602_LINE_WIDTH) return;
	p->framebuf[y][1 + x] = ch;
}

/**
 * Get key from the device.
 * \param drvthis  Pointer to driver structure.
 * \return         String representation of the key;
 *                 \c NULL if nothing available / unmapped key.
 */
MODULE_EXPORT const char *
jbc1602_get_key(Driver *drvthis)
{
	PrivateData *p = drvthis->private_data;
	char * key = NULL;

	if (p->input[0] == '\0') {
		fd_set brfdset;
		struct timeval twait;
		int bufused;
		int remspace;

		/* No data, try to read some */
		FD_ZERO(&brfdset);
		FD_SET(p->fd, &brfdset);

		twait.tv_sec = 0;
		twait.tv_usec = 0;

		bufused = strlen(p->input);
		remspace = sizeof(p->input) - bufused - 1;
		if (remspace <= 0) {
			report(RPT_ERR, "%s: Read overflow in JBC1602 getchar", drvthis->name);
			memset(p->input, 0, sizeof(p->input));
			remspace = sizeof(p->input) - 1;
			bufused = 0;
		}

		if (select(p->fd + 1, &brfdset, NULL, NULL, &twait)) {
			int r;

			/* Read as much as possible */
			r = read(p->fd, p->input + bufused, remspace);
			if (r < 0) {
				if (errno != EAGAIN)
					report(RPT_ERR, "%s: Read error in JBC1602 getchar errno=%d", drvthis->name, errno);
			}
		}
	}
	if (p->input[0] != '\0') {
		switch (p->input[0]) {
		case 'u': key = "Up"; break;
		case 'd': key = "Down"; break;
		case 'e': key = "Escape"; break;
		case 's': key = "Enter"; break;
		default:
			report(RPT_ERR, "%s: Invalid character in JBC1602 getchar: {0x%02x}", drvthis->name, p->input[0]);
			break;
		}
		memmove(p->input, p->input + 1, strlen(p->input));
	}
	return key;
}

MODULE_EXPORT int
jbc1602_get_free_chars (Driver *drvthis)
{
        return 8;
}

MODULE_EXPORT void
jbc1602_set_char (Driver *drvthis, int n, unsigned char *dat)
{
	PrivateData *p = drvthis->private_data;
	fd_set bwfdset;
	unsigned char packet[9];
	unsigned char *b = packet;
	unsigned int rem = sizeof(packet);
	int r;

	packet[0] = 0x20 | (n & 7);
	memcpy(packet + 1, dat, 8);
	FD_ZERO(&bwfdset);
	FD_SET(p->fd, &bwfdset);
	while (rem > 0) {
		if (select(p->fd + 1, NULL, &bwfdset, NULL, NULL) >= 0) {
			r = write(p->fd, b, rem);
			if (r > 0) {
				b += r;
				rem -= r;
			} else {
				report(RPT_ERR, "Failed custom character flush on fd=%d", p->fd);
			}
		}
	}
	tcdrain(p->fd);
}


MODULE_EXPORT int
jbc1602_icon(Driver *drvthis, int x, int y, int icon)
{
/*        PrivateData *p = drvthis->private_data;*/

        debug(RPT_DEBUG, "%s: x=%d, y=%d, icon=0x%02X", __FUNCTION__, x, y, icon);

        static unsigned char heart_open[] =
                { b__XXXXX,
                  b__X_X_X,
                  b_______,
                  b_______,
                  b_______,
                  b__X___X,
                  b__XX_XX,
                  b__XXXXX };
        static unsigned char heart_filled[] =
                { b__XXXXX,
                  b__X_X_X,
                  b___X_X_,
                  b___XXX_,
                  b___XXX_,
                  b__X_X_X,
                  b__XX_XX,
                  b__XXXXX };
        static unsigned char arrow_up[] =
                { b____X__,
                  b___XXX_,
                  b__X_X_X,
                  b____X__,
                  b____X__,
                  b____X__,
                  b____X__,
                  b_______ };
        static unsigned char arrow_down[] =
                { b____X__,
                  b____X__,
                  b____X__,
                  b____X__,
                  b__X_X_X,
                  b___XXX_,
                  b____X__,
                  b_______ };
        static unsigned char checkbox_off[] =
                { b_______,
                  b_______,
                  b__XXXXX,
                  b__X___X,
                  b__X___X,
                  b__X___X,
                  b__XXXXX,
                  b_______ };
        static unsigned char checkbox_on[] =
                { b____X__,
                  b____X__,
                  b__XXX_X,
                  b__X_XX_,
                  b__X_X_X,
                  b__X___X,
                  b__XXXXX,
                  b_______ };
        static unsigned char checkbox_gray[] =
                { b_______,
                  b_______,
                  b__XXXXX,
                  b__X_X_X,
                  b__XX_XX,
                  b__X_X_X,
                  b__XXXXX,
                  b_______ };
	static unsigned char selector_left[] =
		{ b_______,
		  b___X___,
		  b___XX__,
		  b___XXX_,
		  b___XX__,
		  b___X___,
		  b_______,
		  b_______ };
	static unsigned char selector_right[] =
		{ b_______,
		  b_____X_,
		  b____XX_,
		  b___XXX_,
		  b____XX_,
		  b_____X_,
		  b_______,
		  b_______ };

	switch (icon) {
	case ICON_BLOCK_FILLED:
		jbc1602_chr(drvthis, x, y, 0xff);
		break;
	case ICON_HEART_FILLED:
		jbc1602_set_char(drvthis, 0, heart_filled);
		jbc1602_chr(drvthis, x, y, 0);
		break;
	case ICON_HEART_OPEN:
		jbc1602_set_char(drvthis, 0, heart_open);
		jbc1602_chr(drvthis, x, y, 0);
		break;
	case ICON_ARROW_UP:
		jbc1602_set_char(drvthis, 1, arrow_up);
		jbc1602_chr(drvthis, x, y, 1);
		break;
	case ICON_ARROW_DOWN:
		jbc1602_set_char(drvthis, 2, arrow_down);
		jbc1602_chr(drvthis, x, y, 2);
		break;
	case ICON_ARROW_LEFT:
		jbc1602_chr(drvthis, x, y, 0x7F);
		break;
	case ICON_ARROW_RIGHT:
		jbc1602_chr(drvthis, x, y, 0x7E);
		break;
	case ICON_CHECKBOX_OFF:
		jbc1602_set_char(drvthis, 3, checkbox_off);
		jbc1602_chr(drvthis, x, y, 3);
		break;
	case ICON_CHECKBOX_ON:
		jbc1602_set_char(drvthis, 4, checkbox_on);
		jbc1602_chr(drvthis, x, y, 4);
		break;
	case ICON_CHECKBOX_GRAY:
		jbc1602_set_char(drvthis, 5, checkbox_gray);
		jbc1602_chr(drvthis, x, y, 5);
		break;
	case ICON_SELECTOR_AT_LEFT:
		jbc1602_set_char(drvthis, 6, selector_left);
		jbc1602_chr(drvthis, x, y, 6);
		break;
	case ICON_SELECTOR_AT_RIGHT:
		jbc1602_set_char(drvthis, 7, selector_right);
		jbc1602_chr(drvthis, x, y, 7);
		break;
	default:
		return -1;
	}

	return 0;
}

MODULE_EXPORT void
jbc1602_backlight (Driver *drvthis, int on)
{
        PrivateData *p = drvthis->private_data;
        unsigned char hardware_value = (on == BACKLIGHT_ON) ? 0x31 : 0x30;
	fd_set bwfdset;
	int r;

	FD_ZERO(&bwfdset);
	FD_SET(p->fd, &bwfdset);

	if (select(p->fd + 1, NULL, &bwfdset, NULL, NULL) >= 0) {
		r = write(p->fd, &hardware_value, 1);
		if (r <= 0) {
			report(RPT_ERR, "Failed backlight flush on fd=%d", p->fd);
                }
	}
        tcdrain(p->fd);
}

